package kasapp

import (
	"context"
	"fmt"
	"log/slog"
	"math"
	"net"
	"net/netip"
	"os"
	"strconv"
	"sync"
	"time"

	"github.com/ash2k/stager"
	"github.com/golang-jwt/jwt/v5"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/api"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/module/modshared"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/errz"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/grpctool"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/ioz"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/logz"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/nettool"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/prototool"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/tlstool"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/pkg/kascfg"
	"go.opentelemetry.io/contrib/instrumentation/google.golang.org/grpc/otelgrpc"
	"google.golang.org/grpc"
	"google.golang.org/grpc/backoff"
	"google.golang.org/grpc/credentials"
	"google.golang.org/grpc/credentials/insecure"
	"google.golang.org/grpc/keepalive"
)

const (
	envVarOwnPrivateAPIURL    = "OWN_PRIVATE_API_URL"
	envVarOwnPrivateAPICIDR   = "OWN_PRIVATE_API_CIDR"
	envVarOwnPrivateAPIScheme = "OWN_PRIVATE_API_SCHEME"
	envVarOwnPrivateAPIPort   = "OWN_PRIVATE_API_PORT"
	envVarOwnPrivateAPIHost   = "OWN_PRIVATE_API_HOST"
)

type privateAPIServer struct {
	log           *slog.Logger
	listenCfg     *kascfg.ListenPrivateApiCF
	listenMetric  *nettool.ListenerMetrics
	ownURL        string
	server        grpctool.GRPCServer
	listenServer  *grpc.Server
	inMemServer   *grpc.Server
	inMemListener net.Listener
	kasPool       grpctool.PoolInterface
	kasPoolClose  func()
	auxCancel     context.CancelFunc
	ready         func()
}

func newPrivateAPIServer(log *slog.Logger, ot *obsTools, errRep errz.ErrReporter, cfg *kascfg.ConfigurationFile,
	factory modshared.RPCAPIFactory, grpcServerErrorReporter grpctool.ServerErrorReporter, v prototool.Validator,
	listenMetric *nettool.ListenerMetrics) (*privateAPIServer, error) {
	listenCfg := cfg.PrivateApi.Listen
	jwtSecret, err := ioz.LoadSHA256Base64Secret(log, listenCfg.AuthenticationSecretFile)
	if err != nil {
		return nil, fmt.Errorf("auth secret file: %w", err)
	}

	ownURL, err := constructOwnURL(
		net.InterfaceAddrs,
		os.Getenv(envVarOwnPrivateAPIURL),
		os.Getenv(envVarOwnPrivateAPICIDR),
		os.Getenv(envVarOwnPrivateAPIScheme),
		os.Getenv(envVarOwnPrivateAPIPort),
		*listenCfg.Network,
		listenCfg.Address,
	)
	if err != nil {
		return nil, fmt.Errorf("construct own private API multi URL: %w", err)
	}
	log.Info("Using own private API multi URL", logz.URL(ownURL))

	ownHost := os.Getenv(envVarOwnPrivateAPIHost)

	// In-memory gRPC client->listener pipe
	listener := grpctool.NewDialListener()

	// Client pool
	kasPool, err := newKASPool(log, ot, errRep, jwtSecret, ownURL, ownHost, listenCfg.CaCertificateFile,
		v, listener.DialContext)
	if err != nil {
		return nil, fmt.Errorf("kas pool: %w", err)
	}

	// Server
	auxCtx, auxCancel := context.WithCancel(context.Background()) //nolint: govet
	server, inMemServer, err := newPrivateAPIServerImpl(log, auxCtx, ot, cfg, jwtSecret, factory, ownHost, grpcServerErrorReporter, v)
	if err != nil {
		return nil, fmt.Errorf("new server: %w", err) //nolint: govet
	}
	return &privateAPIServer{
		log:           log,
		listenCfg:     listenCfg,
		listenMetric:  listenMetric,
		ownURL:        ownURL,
		server:        grpctool.AggregateServer{server, inMemServer},
		listenServer:  server,
		inMemServer:   inMemServer,
		inMemListener: listener,
		kasPool:       kasPool,
		kasPoolClose:  sync.OnceFunc(func() { kasPool.Shutdown(30 * time.Second) }),
		auxCancel:     auxCancel,
		ready:         ot.probeRegistry.RegisterReadinessToggle("privateAPIServer"),
	}, nil
}

func (s *privateAPIServer) Start(stage stager.Stage) {
	stopInMem := make(chan struct{})
	grpctool.StartServer(stage, s.inMemServer,
		func() (net.Listener, error) {
			return s.inMemListener, nil
		},
		func() {
			<-stopInMem
			s.kasPoolClose()
		},
		func() {},
	)
	grpctool.StartServer(stage, s.listenServer,
		func() (net.Listener, error) {
			lis, err := nettool.ListenWithOSTCPKeepAlive(*s.listenCfg.Network, s.listenCfg.Address)
			if err != nil {
				return nil, err
			}
			lisWrapped, err := s.listenMetric.Wrap(lis, "private_api", s.listenCfg.MaxConnectionAge.AsDuration())
			if err != nil {
				_ = lis.Close()
				return nil, err
			}
			addr := lisWrapped.Addr()
			s.log.Info("Private API endpoint is up",
				logz.NetNetworkFromAddr(addr),
				logz.NetAddressFromAddr(addr),
			)
			s.ready()
			return lisWrapped, nil
		},
		func() {
			time.Sleep(s.listenCfg.ListenGracePeriod.AsDuration())
			// We first want gRPC server to send GOAWAY and only then return from the RPC handlers.
			// So we delay signaling the handlers.
			// See https://github.com/grpc/grpc-go/issues/6830 for more background.
			// Start a goroutine in a second and...
			time.AfterFunc(time.Second, func() {
				close(stopInMem) // ... signal the in-memory server to stop.
				s.auxCancel()    // ... signal running RPC handlers to stop.
			})
		},
		func() {},
	)
}

func (s *privateAPIServer) Close() error {
	s.kasPoolClose()               // first close the client (if not closed already)
	return s.inMemListener.Close() // then close the listener (if not closed already)
}

func newPrivateAPIServerImpl(log *slog.Logger, auxCtx context.Context, ot *obsTools, cfg *kascfg.ConfigurationFile,
	jwtSecret []byte, factory modshared.RPCAPIFactory, ownPrivateAPIHost string,
	grpcServerErrorReporter grpctool.ServerErrorReporter, v prototool.Validator) (*grpc.Server, *grpc.Server, error) {
	listenCfg := cfg.PrivateApi.Listen
	credsOpt, err := grpctool.MaybeTLSCreds(listenCfg.CertificateFile, listenCfg.KeyFile)
	if err != nil {
		return nil, nil, err
	}
	if ownPrivateAPIHost == "" && len(credsOpt) > 0 {
		log.Info(fmt.Sprintf("%s environment variable is not set. Please set it if you want to "+
			"override the server name used for KAS->KAS TLS communication", envVarOwnPrivateAPIHost))
	}

	jwtAuther := grpctool.NewHMACJWTAuther(jwtSecret, api.JWTKAS, api.JWTKAS, func(ctx context.Context) *slog.Logger {
		return modshared.RPCAPIFromContext(ctx).Log()
	})

	keepaliveOpt, sh := grpctool.MaxConnectionAge2GRPCKeepalive(auxCtx, listenCfg.MaxConnectionAge.AsDuration())
	sharedOpts := []grpc.ServerOption{
		keepaliveOpt,
		grpc.StatsHandler(otelgrpc.NewServerHandler(
			otelgrpc.WithTracerProvider(ot.MaybeTraceProvider(ot.grpcServerTracing)),
			otelgrpc.WithMeterProvider(ot.mp),
			otelgrpc.WithPropagators(ot.p),
			otelgrpc.WithMessageEvents(otelgrpc.ReceivedEvents, otelgrpc.SentEvents),
		)),
		grpc.StatsHandler(ot.ssh),
		grpc.StatsHandler(sh),
		grpc.SharedWriteBuffer(true),
		grpc.ChainStreamInterceptor(
			ot.streamProm, // 1. measure all invocations
			modshared.StreamRPCAPIInterceptor(factory),                             // 2. inject RPC API
			jwtAuther.StreamServerInterceptor,                                      // 3. auth and maybe log
			grpctool.StreamServerValidatingInterceptor(v),                          // x. wrap with validator
			grpctool.StreamServerErrorReporterInterceptor(grpcServerErrorReporter), //nolint:contextcheck
		),
		grpc.ChainUnaryInterceptor(
			ot.unaryProm, // 1. measure all invocations
			modshared.UnaryRPCAPIInterceptor(factory),    // 2. inject RPC API
			jwtAuther.UnaryServerInterceptor,             // 3. auth and maybe log
			grpctool.UnaryServerValidatingInterceptor(v), // x. wrap with validator
			grpctool.UnaryServerErrorReporterInterceptor(grpcServerErrorReporter),
		),
		grpc.KeepaliveEnforcementPolicy(keepalive.EnforcementPolicy{
			MinTime:             20 * time.Second,
			PermitWithoutStream: true,
		}),
		grpc.ForceServerCodecV2(grpctool.RawCodecWithProtoFallback{}),
		grpc.MaxRecvMsgSize(api.GRPCMaxMessageSize),
	}
	server := grpc.NewServer(append(sharedOpts, credsOpt...)...)
	inMemServer := grpc.NewServer(sharedOpts...)
	return server, inMemServer, nil
}

func newKASPool(log *slog.Logger, ot *obsTools, errRep errz.ErrReporter, jwtSecret []byte,
	ownPrivateAPIURL, ownPrivateAPIHost, caCertificateFile string, v prototool.Validator,
	dialer func(context.Context, string) (net.Conn, error)) (grpctool.PoolInterface, error) {

	sharedPoolOpts := []grpc.DialOption{
		grpc.WithSharedWriteBuffer(true),
		// Default gRPC parameters are good, no need to change them at the moment.
		// Specify them explicitly for discoverability.
		// See https://github.com/grpc/grpc/blob/master/doc/connection-backoff.md.
		grpc.WithConnectParams(grpc.ConnectParams{
			Backoff:           backoff.DefaultConfig,
			MinConnectTimeout: 20 * time.Second, // matches the default gRPC value.
		}),
		grpc.WithStatsHandler(otelgrpc.NewClientHandler(
			otelgrpc.WithTracerProvider(ot.MaybeTraceProvider(ot.grpcClientTracing)),
			otelgrpc.WithMeterProvider(ot.mp),
			otelgrpc.WithPropagators(ot.p),
			otelgrpc.WithMessageEvents(otelgrpc.ReceivedEvents, otelgrpc.SentEvents),
		)),
		grpc.WithStatsHandler(ot.csh),
		grpc.WithUserAgent(kasServerName()),
		grpc.WithKeepaliveParams(keepalive.ClientParameters{
			Time:                55 * time.Second,
			PermitWithoutStream: true,
		}),
		grpc.WithPerRPCCredentials(&grpctool.JWTCredentials{
			SigningMethod: jwt.SigningMethodHS256,
			SigningKey:    jwtSecret,
			Audience:      api.JWTKAS,
			Issuer:        api.JWTKAS,
			Insecure:      true, // We may or may not have TLS setup, so always say creds don't need TLS.
		}),
		grpc.WithChainStreamInterceptor(
			ot.streamClientProm,
			grpctool.StreamClientValidatingInterceptor(v),
		),
		grpc.WithChainUnaryInterceptor(
			ot.unaryClientProm,
			grpctool.UnaryClientValidatingInterceptor(v),
		),
		grpc.WithNoProxy(),
		grpc.WithDefaultCallOptions(grpc.MaxCallRecvMsgSize(api.GRPCMaxMessageSize)),
	}

	// Construct in-memory connection to private API gRPC server
	inMemConn, err := grpc.NewClient("passthrough:private-server",
		append(sharedPoolOpts,
			grpc.WithContextDialer(dialer),
			grpc.WithTransportCredentials(insecure.NewCredentials()),
		)...,
	)
	if err != nil {
		return nil, err
	}
	tlsConfig, err := tlstool.ClientConfigWithCACert(caCertificateFile)
	if err != nil {
		return nil, err
	}
	tlsConfig.ServerName = ownPrivateAPIHost
	tlsCredsOpt := grpc.WithTransportCredentials(credentials.NewTLS(tlsConfig))
	insecureCredsOpt := grpc.WithTransportCredentials(insecure.NewCredentials())
	kasPool := grpctool.NewPool(log, errRep, func(targetURL string) (string, []grpc.DialOption, error) {
		mu, err := nettool.ParseMultiURL(targetURL)
		if err != nil {
			return "", nil, err
		}
		var credsOpt grpc.DialOption
		switch mu.Scheme() {
		case "grpc":
			credsOpt = insecureCredsOpt
		case "grpcs":
			credsOpt = tlsCredsOpt
		default:
			return "", nil, fmt.Errorf("unsupported URL scheme in %s", targetURL)
		}
		l := len(sharedPoolOpts)
		muOpts := append(
			sharedPoolOpts[:l:l], // set cap=len to ensure append() will copy the slice
			grpc.WithResolvers(grpctool.NewMultiURLResolver(mu)),
			credsOpt,
		)
		return grpctool.MultiURLResolverTarget, muOpts, nil
	})
	return grpctool.NewPoolSelf(log, errRep, kasPool, ownPrivateAPIURL, inMemConn), nil
}

func constructOwnURL(interfaceAddrs func() ([]net.Addr, error),
	ownURL, ownCIDR, ownScheme, ownPort, listenNetwork, listenAddress string) (string, error) {

	if ownURL != "" {
		if ownCIDR != "" {
			return "", fmt.Errorf("either %s or %s should be specified, not both", envVarOwnPrivateAPIURL, envVarOwnPrivateAPICIDR)
		}
		u, err := nettool.ParseURL(ownURL)
		if err != nil {
			return "", fmt.Errorf("failed to parse %s %s: %w", envVarOwnPrivateAPIURL, ownURL, err)
		}
		return u.String(), err
	}

	// Determine port. 0 means not set
	port, err := detectOwnPort(ownPort)
	if err != nil {
		return "", err
	}

	// Determine scheme
	scheme, err := detectOwnScheme(ownScheme)
	if err != nil {
		return "", err
	}

	if ownCIDR != "" {
		return detectURLByCIDR(interfaceAddrs, ownCIDR, scheme, port, listenNetwork, listenAddress)
	}

	return detectURLFromListenAddress(interfaceAddrs, scheme, port, listenNetwork, listenAddress)
}

func detectOwnScheme(ownScheme string) (string, error) {
	switch ownScheme {
	case "grpc", "grpcs":
		return ownScheme, nil
	case "":
		return "grpc", nil
	default:
		return "", fmt.Errorf("%s environment variable should be either grpc or grpcs, got: %s", envVarOwnPrivateAPIScheme, ownScheme)
	}
}

func detectOwnPort(ownPort string) (uint16, error) {
	if ownPort == "" {
		return 0, nil
	}
	port, err := parsePort(ownPort)
	if err != nil {
		return 0, fmt.Errorf("error parsing %s environment variable: %w", envVarOwnPrivateAPIPort, err)
	}
	return port, nil
}

func detectURLByCIDR(interfaceAddrs func() ([]net.Addr, error),
	ownCIDR, scheme string, port uint16, listenNetwork, listenAddress string) (string, error) {
	_, ipNet, err := net.ParseCIDR(ownCIDR)
	if err != nil {
		return "", fmt.Errorf("failed to parse %s environment variable: %w", envVarOwnPrivateAPICIDR, err)
	}
	addrs, err := interfaceAddrs()
	if err != nil {
		return "", fmt.Errorf("net.InterfaceAddrs(): %w", err)
	}
	var foundIPs []netip.Addr //nolint:prealloc
	for _, addr := range addrs {
		addrIP, ok := addr.(*net.IPNet)
		if !ok {
			continue
		}
		if !ipNet.Contains(addrIP.IP) {
			continue
		}
		ip, valid := nettool.IP2Addr(addrIP.IP)
		if !valid {
			// this shouldn't ever happen
			return "", fmt.Errorf("error parsing IP address %s", addrIP.IP)
		}
		foundIPs = append(foundIPs, ip)
	}
	if len(foundIPs) == 0 {
		return "", fmt.Errorf("no IPs matched CIDR specified in the %s environment variable", envVarOwnPrivateAPICIDR)
	}
	if port == 0 { // not specified, use port from listener
		switch listenNetwork {
		case "tcp", "tcp4", "tcp6": // assume listenAddress is a ip:port or name:port
			_, listenPort, err := net.SplitHostPort(listenAddress)
			if err != nil {
				return "", fmt.Errorf("listener address: %w", err)
			}
			port, err = parsePort(listenPort)
			if err != nil {
				return "", fmt.Errorf("listener address port: %w", err)
			}
		//case "unix": We don't handle unix scheme here because we got OWN_PRIVATE_API_CIDR and hence presumably
		// user wants to use network, not unix socket.
		default:
			return "", fmt.Errorf("cannot determine port. Specify %s", envVarOwnPrivateAPIPort)
		}
	}
	return nettool.NewMultiURL(scheme, foundIPs, port).String(), nil
}

func detectURLFromListenAddress(interfaceAddrs func() ([]net.Addr, error),
	scheme string, port uint16, listenNetwork, listenAddress string) (string, error) {
	switch listenNetwork {
	case "tcp", "tcp4", "tcp6": // assume listenAddress is a ip:port or name:port
		listenHost, listenPort, err := net.SplitHostPort(listenAddress)
		if err != nil {
			return "", fmt.Errorf("listener address: %w", err)
		}
		if port == 0 {
			// parse listenPort
			port, err = parsePort(listenPort)
			if err != nil {
				return "", fmt.Errorf("listener address port: %w", err)
			}
		}
		if listenHost == "" { // all IPs
			lTCP := listenNetwork == "tcp"
			return multiURLForInterfaceIPs(interfaceAddrs, scheme, port, lTCP || listenNetwork == "tcp4", lTCP || listenNetwork == "tcp6")
		}
		ip, err := netip.ParseAddr(listenHost)
		if err == nil && ip.IsUnspecified() { // all v4 or v6 IPs - '0.0.0.0' or '::'
			return multiURLForInterfaceIPs(interfaceAddrs, scheme, port, ip == netip.IPv4Unspecified(), ip == netip.IPv6Unspecified())
		}
		if err != nil || ip.Zone() != "" || ip.IsLinkLocalUnicast() || ip.IsLoopback() {
			// Not an acceptable IP address. Just report an error.
			return "", fmt.Errorf(
				"cannot determine URL. Please set %s or %s environment variable or change private API listen address to listen on all interfaces (currently set to %q)",
				envVarOwnPrivateAPIURL, envVarOwnPrivateAPICIDR, listenAddress)
		}
		return nettool.NewMultiURL(scheme, []netip.Addr{ip}, port).String(), nil
	default:
		return "", fmt.Errorf("unsupported network type specified in the listener config: %s", listenNetwork)
	}
}

func multiURLForInterfaceIPs(interfaceAddrs func() ([]net.Addr, error), scheme string, port uint16, addV4, addV6 bool) (string, error) {
	addrs, err := interfaceAddrs()
	if err != nil {
		return "", fmt.Errorf("net.InterfaceAddrs(): %w", err)
	}
	var preferredIPs []netip.Addr
	for _, addr := range addrs {
		addrIP, ok := addr.(*net.IPNet)
		if !ok {
			continue
		}
		ip, valid := nettool.IP2Addr(addrIP.IP)
		if !valid {
			// this shouldn't ever happen
			return "", fmt.Errorf("error parsing IP address %s", addrIP.IP)
		}
		if ip.IsLinkLocalUnicast() || ip.IsLoopback() {
			// We assume link-local and loopback addresses are never used in production in multi-node kas setup.
			// If you want to use them (e.g. in a development setup), explicitly set OWN_PRIVATE_API_URL.
			// We could allow such addresses to simplify development setup, but that may lead to
			// kas not detecting misconfiguration when run in a bad production environment
			// (without a proper IP address).
			continue
		}
		if (addV4 && ip.Is4()) || (addV6 && ip.Is6()) {
			preferredIPs = append(preferredIPs, ip)
		}
	}
	if len(preferredIPs) == 0 {
		return "", fmt.Errorf("cannot determine URL - no matching network interfaces found. Please set %s environment variable", envVarOwnPrivateAPIURL)
	}
	return nettool.NewMultiURL(scheme, preferredIPs, port).String(), nil
}

func parsePort(portStr string) (uint16, error) {
	port, err := strconv.ParseUint(portStr, 10, 16)
	if err != nil {
		return 0, fmt.Errorf("parsing port: %w", err)
	}
	if port == 0 || port > math.MaxUint16 { // mostly to check for 0, but do a full range check for completeness.
		return 0, fmt.Errorf("invalid port: %d", port)
	}
	return uint16(port), nil
}
