package router

import (
	"context"
	"errors"
	"net"
	"net/url"
	"strconv"
	"time"

	"github.com/redis/rueidis"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/nettool"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/redistool"
	otelmetric "go.opentelemetry.io/otel/metric"
	"k8s.io/apimachinery/pkg/util/sets"
	"k8s.io/utils/clock"
)

const (
	legacyTunnelsByAgentIDHashName = "legacy_tunnels_by_agent_id"
	tunnelsByAgentIDHashName       = "tunnels_by_agent_id"
)

type Querier interface {
	// KASURLsByAgentID returns the list of kas URLs for a particular agent id.
	// A partial list may be returned together with an error.
	// Safe for concurrent use.
	KASURLsByAgentID(ctx context.Context, agentID int64) ([]string, error)
}

// RegistrationBuilder allows batching tunnel (un)registrations.
// Can be reused after Do is called.
type RegistrationBuilder interface {
	// Register registers tunnels for the given agent ids with the tracker.
	Register(ttl time.Duration, agentIDs ...int64)
	// Unregister unregisters tunnels for the given agent ids with the tracker.
	Unregister(agentIDs ...int64)
	// Refresh refreshes registered tunnels in the underlying storage.
	Refresh(ttl time.Duration, agentIDs ...int64)
	// Do executes the enqueued operations.
	Do(context.Context) error
}

// Registerer allows to register and unregister tunnels.
// Caller is responsible for periodically calling GC() and Refresh().
// Not safe for concurrent use.
type Registerer interface {
	RegistrationBuilder() RegistrationBuilder
	// GC deletes expired tunnels from the underlying storage.
	GC(ctx context.Context, agentIDs []int64) (int /* keysDeleted */, error)
}

type Tracker interface {
	Registerer
	Querier
}

type RedisTracker struct {
	legacyOwnPrivateAPIURL string
	ownPrivateAPIURL       string
	clock                  clock.PassiveClock
	legacyTunnelsByAgentID redistool.ExpiringHashAPI[int64, string] // agentID -> kas URL -> nil
	tunnelsByAgentID       redistool.ExpiringHashAPI[int64, string] // agentID -> kas multi URL -> nil
}

func NewRedisTracker(client rueidis.Client, agentKeyPrefix string, ownPrivateAPIURL string, m otelmetric.Meter) (*RedisTracker, error) {
	legacyTunnelsByAgentID, err := redistool.NewRedisExpiringHashAPI[int64, string](legacyTunnelsByAgentIDHashName, client, legacyTunnelsByAgentIDHashKey(agentKeyPrefix), strToStr, m)
	if err != nil {
		return nil, err
	}
	tunnelsByAgentID, err := redistool.NewRedisExpiringHashAPI[int64, string](tunnelsByAgentIDHashName, client, tunnelsByAgentIDHashKey(agentKeyPrefix), strToStr, m)
	if err != nil {
		return nil, err
	}
	mu, err := nettool.ParseMultiURL(ownPrivateAPIURL)
	if err != nil {
		return nil, err
	}
	return &RedisTracker{
		legacyOwnPrivateAPIURL: (&url.URL{
			Scheme: mu.Scheme(),
			Host:   net.JoinHostPort(mu.Addresses()[0].String(), strconv.FormatUint(uint64(mu.Port()), 10)),
		}).String(),
		ownPrivateAPIURL:       ownPrivateAPIURL,
		clock:                  clock.RealClock{},
		legacyTunnelsByAgentID: legacyTunnelsByAgentID,
		tunnelsByAgentID:       tunnelsByAgentID,
	}, nil
}

func (t *RedisTracker) RegistrationBuilder() RegistrationBuilder {
	return &redisRegistrationBuilder{
		legacyOwnPrivateAPIURL: t.legacyOwnPrivateAPIURL,
		ownPrivateAPIURL:       t.ownPrivateAPIURL,
		clock:                  t.clock,
		legacyIOB:              t.legacyTunnelsByAgentID.IOBuilder(),
		iob:                    t.tunnelsByAgentID.IOBuilder(),
	}
}

func (t *RedisTracker) KASURLsByAgentID(ctx context.Context, agentID int64) ([]string, error) {
	urls := make(sets.Set[string])
	var errs []error
	err := t.legacyTunnelsByAgentID.Scan(ctx, agentID, func(rawHashKey string, value []byte, err error) (bool, error) {
		if err != nil {
			errs = append(errs, err)
			return false, nil
		}
		if rawHashKey == t.legacyOwnPrivateAPIURL { // founds our own legacy URL
			// skip it since we'll get the full one in the Scan() below
			return false, nil
		}
		mu, err := nettool.ParseURL(rawHashKey) // turn URL into a multi URL
		if err != nil {
			errs = append(errs, err)
			return false, nil
		}
		urls.Insert(mu.String())
		return false, nil
	})
	if err != nil {
		errs = append(errs, err)
	}
	err = t.tunnelsByAgentID.Scan(ctx, agentID, func(rawHashKey string, value []byte, err error) (bool, error) {
		if err != nil {
			errs = append(errs, err)
			return false, nil
		}
		urls.Insert(rawHashKey)
		return false, nil
	})
	if err != nil {
		errs = append(errs, err)
	}
	return urls.UnsortedList(), errors.Join(errs...)
}

func (t *RedisTracker) GC(ctx context.Context, agentIDs []int64) (int /* keysDeleted */, error) {
	d1, err1 := t.legacyTunnelsByAgentID.GCFor(agentIDs, true)(ctx)
	d2, err2 := t.tunnelsByAgentID.GCFor(agentIDs, true)(ctx)
	return d1 + d2, errors.Join(err1, err2)
}

type redisRegistrationBuilder struct {
	legacyOwnPrivateAPIURL string
	ownPrivateAPIURL       string
	clock                  clock.PassiveClock
	legacyIOB              redistool.IOBuilder[int64, string]
	iob                    redistool.IOBuilder[int64, string]
}

func (b *redisRegistrationBuilder) Register(ttl time.Duration, agentIDs ...int64) {
	exp := b.clock.Now().Add(ttl)
	b.legacyIOB.Set(agentIDs, ttl, kv(b.legacyOwnPrivateAPIURL, exp))
	b.iob.Set(agentIDs, ttl, kv(b.ownPrivateAPIURL, exp))
}

func (b *redisRegistrationBuilder) Unregister(agentIDs ...int64) {
	b.legacyIOB.Unset(agentIDs, b.legacyOwnPrivateAPIURL)
	b.iob.Unset(agentIDs, b.ownPrivateAPIURL)
}

func (b *redisRegistrationBuilder) Refresh(ttl time.Duration, agentIDs ...int64) {
	// Refresh is the same as registration right now, but this might change in the future.
	b.Register(ttl, agentIDs...)
}

func (b *redisRegistrationBuilder) Do(ctx context.Context) error {
	return errors.Join(
		b.legacyIOB.Do(ctx),
		b.iob.Do(ctx),
	)
}

func kv(key string, expiresAt time.Time) redistool.BuilderKV[string] {
	return redistool.BuilderKV[string]{
		HashKey: key,
		Value: &redistool.ExpiringValue{
			ExpiresAt: expiresAt.Unix(),
			Value:     nil, // nothing to store.
		},
	}
}

// legacyTunnelsByAgentIDHashKey returns a key for agentID -> (kas URL -> nil).
func legacyTunnelsByAgentIDHashKey(agentKeyPrefix string) redistool.KeyToRedisKey[int64] {
	prefix := agentKeyPrefix + ":kas_by_agent_id:"
	return func(agentID int64) string {
		return redistool.PrefixedInt64Key(prefix, agentID)
	}
}

// tunnelsByAgentIDHashKey returns a key for agentID -> (kas multi URL -> nil).
func tunnelsByAgentIDHashKey(agentKeyPrefix string) redistool.KeyToRedisKey[int64] {
	prefix := agentKeyPrefix + ":kas_by_agent_id2:"
	return func(agentID int64) string {
		return redistool.PrefixedInt64Key(prefix, agentID)
	}
}

func strToStr(key string) string {
	return key
}
