package nettool

import (
	"fmt"
	"net/netip"
	"net/url"
	"slices"
	"strconv"
	"strings"
)

// MultiURL is an immutable URL that has scheme, port, and one or more IP addresses.
type MultiURL struct {
	scheme    string
	addresses []netip.Addr
	port      uint16
}

func NewMultiURL(scheme string, addresses []netip.Addr, port uint16) MultiURL {
	return MultiURL{
		scheme:    scheme,
		addresses: slices.Clone(addresses),
		port:      port,
	}
}

func (u MultiURL) String() string {
	if len(u.addresses) == 0 {
		return "invalid multi address: no addresses"
	}
	if u.scheme == "" {
		return "invalid multi address: empty scheme"
	}
	var sb strings.Builder
	sb.WriteString(u.scheme)
	sb.WriteByte(':')
	for i, addr := range u.addresses {
		if i > 0 {
			sb.WriteByte(',')
		}
		sb.WriteString(addr.String())
	}
	sb.WriteByte('|')
	sb.WriteString(strconv.FormatUint(uint64(u.port), 10))
	return sb.String()
}

func (u MultiURL) Scheme() string {
	return u.scheme
}

func (u MultiURL) Addresses() []netip.Addr {
	return slices.Clone(u.addresses) // do not expose the underlying slice.
}

func (u MultiURL) Port() uint16 {
	return u.port
}

func IsEqualMultiURL(a, b MultiURL) bool {
	return a.scheme == b.scheme && a.port == b.port && slices.Equal(a.addresses, b.addresses)
}

// ParseMultiURL parses addr according to the following opaque URL schema:
// scheme:ip[,ip]|port
func ParseMultiURL(addr string) (MultiURL, error) {
	scheme, rest, found := strings.Cut(addr, ":")
	if !found {
		return MultiURL{}, fmt.Errorf("scheme not found in: %s", addr)
	}
	ipsStr, portStr, found := strings.Cut(rest, "|")
	if !found {
		return MultiURL{}, fmt.Errorf("port not found in: %s", addr)
	}
	port, err := strconv.ParseUint(portStr, 10, 16)
	if err != nil {
		return MultiURL{}, fmt.Errorf("port in %s: %w", addr, err)
	}
	ipsList := strings.Split(ipsStr, ",")
	if len(ipsList) == 0 {
		return MultiURL{}, fmt.Errorf("no IPs in: %s", addr)
	}
	addresses := make([]netip.Addr, 0, len(ipsList))
	for _, ipStr := range ipsList {
		ip, err := netip.ParseAddr(ipStr)
		if err != nil {
			return MultiURL{}, fmt.Errorf("IP parsing in %s: %w", addr, err)
		}
		addresses = append(addresses, ip)
	}
	return MultiURL{
		scheme:    scheme,
		addresses: addresses,
		port:      uint16(port),
	}, nil
}

func ParseURL(addr string) (MultiURL, error) {
	u, err := url.Parse(addr)
	if err != nil {
		return MultiURL{}, err
	}
	return MultiURLFromURL(u)
}

// MultiURLFromURL turns a URL into a MultiURL.
// The provided URL must:
// - not be an opaque.
// - have a scheme.
// - have a port.
func MultiURLFromURL(u *url.URL) (MultiURL, error) {
	if u.Scheme == "" {
		return MultiURL{}, fmt.Errorf("no scheme in: %s", u)
	}
	addrPort, err := netip.ParseAddrPort(u.Host)
	if err != nil {
		return MultiURL{}, err
	}
	return MultiURL{
		scheme:    u.Scheme,
		addresses: []netip.Addr{addrPort.Addr()},
		port:      addrPort.Port(),
	}, nil
}
