package it

import (
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/gitaly/vendored/gitalypb"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/module/managed_resources"
	managed_resources_rpc "gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/module/managed_resources/rpc"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/testing/matcher"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/testing/mock_gitaly"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/pkg/agentcfg"
	"go.uber.org/mock/gomock"
	"google.golang.org/grpc"
)

func (s *integrationSuite) TestManagedResources() {
	gitalySrv := s.startGitaly(gitalyOptions{})
	s.setupGitLabMocks(gitalySrv, true)
	s.setupGitalyMocksForAgentConfig(&agentcfg.ConfigurationFile{
		Observability: &agentcfg.ObservabilityCF{
			Logging: &agentcfg.LoggingCF{
				Level:     agentcfg.LogLevelEnum_debug,
				GrpcLevel: ptr(agentcfg.LogLevelEnum_debug),
			},
		},
	})
	gitLabSrv := s.startGitLab(gitLabOptions{})

	kas := s.startKAS(kasOptions{
		name:       "kas",
		cfg:        s.baseKASConfig(gitLabSrv),
		gitalyPort: gitalySrv.port,
	})
	s.startAgentk(agentkOptions{
		kasURL: kas.agentServerURL,
		token:  agentkToken1,
		name:   "agent1",
		ns:     s.ns,
	})

	kasAPI := s.clientForKASAPI(kas)
	c := managed_resources_rpc.NewProvisionerClient(kasAPI)
	info := templatingInfo()
	gitalyInfo, gitalyRepository := gitalyFields(gitalySrv)

	s.Run("Render the default template", func() {
		tpl, err := c.GetDefaultEnvironmentTemplate(s.ctx, &managed_resources_rpc.GetDefaultEnvironmentTemplateRequest{})
		s.Require().NoError(err)
		s.T().Log("Template is:\n", string(tpl.Template.Data))

		renderedTpl, err := c.RenderEnvironmentTemplate(s.ctx, &managed_resources_rpc.RenderEnvironmentTemplateRequest{
			Template: tpl.Template,
			Info:     info,
		})
		s.Require().NoError(err)
		s.T().Log("Rendered template is:\n", string(renderedTpl.Template.Data))

		envResp, err := c.EnsureEnvironment(s.ctx, &managed_resources_rpc.EnsureEnvironmentRequest{
			Template: renderedTpl.Template,
			Info:     info,
		})
		s.Require().NoError(err)

		s.assertObjectsExist(envResp.Objects)
		s.Empty(envResp.Errors)

		envResp, err = c.EnsureEnvironment(s.ctx, &managed_resources_rpc.EnsureEnvironmentRequest{
			Template: renderedTpl.Template,
			Info:     info,
		})
		s.Require().NoError(err)
		s.assertObjectsExist(envResp.Objects)
		s.Empty(envResp.Errors)
	})
	s.Run("ListEnvironmentTemplates for all agents", func() {
		s.gitalyM.EXPECT().
			GetTreeEntries(mock_gitaly.GetTreeEntriesRequestForAllAgents(), gomock.Any()).
			DoAndReturn(func(r *gitalypb.GetTreeEntriesRequest, srv grpc.ServerStreamingServer[gitalypb.GetTreeEntriesResponse]) error {
				return srv.Send(&gitalypb.GetTreeEntriesResponse{
					Entries: []*gitalypb.TreeEntry{
						{
							Path: []byte(".gitlab/agents/agent1/environment_templates/name1.yaml"),
							Type: gitalypb.TreeEntry_BLOB,
						},
						{
							Path: []byte(".gitlab/agents/agent2/environment_templates/name2.yaml"),
							Type: gitalypb.TreeEntry_BLOB,
						},
					},
				})
			})
		resp, err := c.ListEnvironmentTemplates(s.ctx, &managed_resources_rpc.ListEnvironmentTemplatesRequest{
			AgentName:        nil, // all agents
			GitalyInfo:       gitalyInfo,
			GitalyRepository: gitalyRepository,
			DefaultBranch:    "main",
		})
		s.Require().NoError(err)

		s.Require().Len(resp.Infos, 2)
		matcher.AssertProtoEqual(s.T(), resp.Infos[0], &managed_resources_rpc.EnvironmentTemplateInfo{AgentName: "agent1", TemplateName: "name1"})
		matcher.AssertProtoEqual(s.T(), resp.Infos[1], &managed_resources_rpc.EnvironmentTemplateInfo{AgentName: "agent2", TemplateName: "name2"})
	})
	s.Run("ListEnvironmentTemplates for a particular agent", func() {
		s.gitalyM.EXPECT().
			GetTreeEntries(mock_gitaly.GetTreeEntriesRequestForAnAgent(agentName), gomock.Any()).
			DoAndReturn(func(r *gitalypb.GetTreeEntriesRequest, srv grpc.ServerStreamingServer[gitalypb.GetTreeEntriesResponse]) error {
				return srv.Send(&gitalypb.GetTreeEntriesResponse{
					Entries: []*gitalypb.TreeEntry{
						{
							Path: []byte(".gitlab/agents/" + agentName + "/environment_templates/name1.yaml"),
							Type: gitalypb.TreeEntry_BLOB,
						},
					},
				})
			})
		resp, err := c.ListEnvironmentTemplates(s.ctx, &managed_resources_rpc.ListEnvironmentTemplatesRequest{
			AgentName:        ptr(agentName),
			GitalyInfo:       gitalyInfo,
			GitalyRepository: gitalyRepository,
			DefaultBranch:    "main",
		})
		s.Require().NoError(err)

		s.Require().Len(resp.Infos, 1)
		matcher.AssertProtoEqual(s.T(), resp.Infos[0], &managed_resources_rpc.EnvironmentTemplateInfo{AgentName: agentName, TemplateName: "name1"})
	})
	s.Run("Get a template for an agent", func() {
		t1Data := []byte("template: n11")
		s.setupTreeEntryMockForTemplate("agent1", "name1", t1Data)
		resp, err := c.GetEnvironmentTemplate(s.ctx, &managed_resources_rpc.GetEnvironmentTemplateRequest{
			TemplateName:     "name1",
			AgentName:        "agent1",
			GitalyInfo:       gitalyInfo,
			GitalyRepository: gitalyRepository,
			DefaultBranch:    "main",
		})
		s.Require().NoError(err)
		matcher.AssertProtoEqual(s.T(), resp.Template, &managed_resources.EnvironmentTemplate{Name: "name1", Data: t1Data})
	})
}

func (s *integrationSuite) setupTreeEntryMockForTemplate(agentName, templateName string, data []byte) {
	s.gitalyM.EXPECT().
		TreeEntry(mock_gitaly.TreeEntryRequestForATemplate(agentName, templateName), gomock.Any()).
		DoAndReturn(func(r *gitalypb.TreeEntryRequest, srv grpc.ServerStreamingServer[gitalypb.TreeEntryResponse]) error {
			return srv.Send(&gitalypb.TreeEntryResponse{
				Type: gitalypb.TreeEntryResponse_BLOB,
				Size: int64(len(data)),
				Data: data,
			})
		})
}
